/*
 * Copyright (c) 2021 Satlab A/S <satlab@satlab.com>
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include <volk/volk.h>
#include "node_sync_impl.h"

namespace gr {
  namespace satlab {

    node_sync::sptr
    node_sync::make(uint64_t timeout_samples)
    {
      return gnuradio::get_initial_sptr
        (new node_sync_impl(timeout_samples));
    }

    void
    node_sync_impl::kick(void)
    {
        d_last_kick = nitems_read(0);
    }

    void
    node_sync_impl::handle_kick(pmt::pmt_t pdu)
    {
        kick();
    }

    int node_sync_impl::general_work(int noutput_items,
                                 gr_vector_int& ninput_items,
                                 gr_vector_const_void_star& input_items,
                                 gr_vector_void_star& output_items)
    {
        float* iptr = (float*)input_items[0];
        float* optr = (float*)output_items[0];

        int nin = std::min(noutput_items, ninput_items[0]);
        int nout = nin;

        if (nitems_read(0) - d_last_kick > d_timeout_samples) {
            kick();
            iptr++;
            nout--;
        }

        std::copy(iptr, iptr+nout, optr);

        consume_each(nin);
        return nout;
    }

    bool node_sync_impl::fixed_rate()
    {
        return false;
    }

    void node_sync_impl::forecast(int noutput_items, gr_vector_int &ninput_items_required)
    {
        ninput_items_required[0] = noutput_items;
    }

    void node_sync_impl::set_timeout(uint64_t samples)
    {
        d_timeout_samples = samples;
    }

    node_sync_impl::node_sync_impl(uint64_t timeout_samples)
      : gr::block("node_sync",
              gr::io_signature::make(1, 1, sizeof(float)),
              gr::io_signature::make(1, 1, sizeof(float))),
        d_last_kick(0),
        d_timeout_samples(timeout_samples)
    {
        message_port_register_in(pmt::mp("kick"));
        set_msg_handler(pmt::mp("kick"), [this](pmt::pmt_t msg) {
                this->handle_kick(msg);
        });
        kick();
    }
  } /* namespace satlab */
} /* namespace gr */
