/*
 * Copyright (c) 2022 Satlab A/S <satlab@satlab.com>
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "snr_est_impl.h"

namespace gr {
    namespace satlab {

        snr_est::sptr
        snr_est::make(digital::constellation_sptr constellation, int navg)
        {
            return gnuradio::get_initial_sptr
                (new snr_est_impl(constellation, navg));
        }

        int snr_est_impl::work(int noutput_items,
                               gr_vector_const_void_star& input_items,
                               gr_vector_void_star& output_items)
        {
            size_t navg = decimation();
            int nin = noutput_items * navg;
            int nout = 0;

            gr_complex* iptr = (gr_complex*)input_items[0];
            std::vector<gr_complex> constl_points = d_constellation->points();

            while (nin >= navg && nout < noutput_items) {
                double d = 0;

                for (int i = 0; i < navg; i++) {
                    // grab best matching constellation point
                    gr_complex k = constl_points[d_constellation->decision_maker(iptr)];

                    // accumulate noise power
                    d += std::pow(std::abs(*iptr - k), 2);

                    // move on to next input symbol
                    iptr++;
                }

                // mean noise power
                d = d / navg;

                // update level: average symbol power to average noise power ratio
                d_level = d_constl_pwr / d;
                nout++;

                consume(0, navg);
                nin -= navg;
            }

            return nout;
        }

        snr_est_impl::snr_est_impl(digital::constellation_sptr constellation, int navg)
            : gr::sync_decimator("snr_est",
                                 gr::io_signature::make(1, 1, sizeof(gr_complex)),
                                 gr::io_signature::make(0, 0, 0),
                                 navg),
              d_constellation(constellation),
              d_level(0)
        {
            // compute average symbol power up front
            d_constl_pwr = 0;
            for (gr_complex c: d_constellation->points())
                d_constl_pwr += std::pow(std::abs(c), 2);
            d_constl_pwr /= d_constellation->points().size();
        }

        double snr_est_impl::level()
        {
            return d_level;
        }

    } /* namespace satlab */
} /* namespace gr */
